@@ -20,7 +20,7 @@ from __future__ import annotations
 import sys
 from typing import Any
 
-from pydantic import JsonValue
+from pydantic import JsonValue, RootModel
 
 from airflow.api_fastapi.core_api.base import BaseModel
 
@@ -36,3 +36,15 @@ class XComResponse(BaseModel):
     key: str
     value: JsonValue
     """The returned XCom value in a JSON-compatible format."""
+
+
+class XComSequenceIndexResponse(RootModel):
+    """XCom schema with minimal structure for index-based access."""
+
+    root: JsonValue
+
+
+class XComSequenceSliceResponse(RootModel):
+    """XCom schema with minimal structure for slice-based access."""
+
+    root: list[JsonValue]
