@@ -31,8 +31,9 @@ import type { PartialEventContext } from "chartjs-plugin-annotation";
 import annotationPlugin from "chartjs-plugin-annotation";
 import dayjs from "dayjs";
 import { Bar } from "react-chartjs-2";
+import { useNavigate } from "react-router-dom";
 
-import type { TaskInstanceResponse, DAGRunResponse } from "openapi/requests/types.gen";
+import type { TaskInstanceResponse, GridRunsResponse } from "openapi/requests/types.gen";
 import { system } from "src/theme";
 import { pluralize } from "src/utils";
 
@@ -53,7 +54,7 @@ const average = (ctx: PartialEventContext, index: number) => {
   return values === undefined ? 0 : values.reduce((initial, next) => initial + next, 0) / values.length;
 };
 
-type RunResponse = DAGRunResponse | TaskInstanceResponse;
+type RunResponse = GridRunsResponse | TaskInstanceResponse;
 
 const getDuration = (start: string, end: string | null) => dayjs.duration(dayjs(end).diff(start)).asSeconds();
 
@@ -64,6 +65,8 @@ export const DurationChart = ({
   readonly entries: Array<RunResponse> | undefined;
   readonly kind: "Dag Run" | "Task Instance";
 }) => {
+  const navigate = useNavigate();
+
   if (!entries) {
     return undefined;
   }
@@ -105,7 +108,7 @@ export const DurationChart = ({
               data: entries.map((entry: RunResponse) => {
                 switch (kind) {
                   case "Dag Run": {
-                    const run = entry as DAGRunResponse;
+                    const run = entry as GridRunsResponse;
 
                     return run.queued_at !== null && run.start_date !== null && run.queued_at < run.start_date
                       ? Number(getDuration(run.queued_at, run.start_date))
@@ -141,6 +144,34 @@ export const DurationChart = ({
         }}
         datasetIdKey="id"
         options={{
+          onClick: (_event, elements) => {
+            const [element] = elements;
+
+            if (!element) {
+              return;
+            }
+
+            switch (kind) {
+              case "Dag Run": {
+                const entry = entries[element.index] as GridRunsResponse | undefined;
+                const baseUrl = `/dags/${entry?.dag_id}/runs/${entry?.run_id}`;
+
+                navigate(baseUrl);
+                break;
+              }
+              case "Task Instance": {
+                const entry = entries[element.index] as TaskInstanceResponse | undefined;
+                const baseUrl = `/dags/${entry?.dag_id}/runs/${entry?.dag_run_id}`;
+
+                navigate(`${baseUrl}/tasks/${entry?.task_id}`);
+                break;
+              }
+              default:
+            }
+          },
+          onHover: (_event, elements, chart) => {
+            chart.canvas.style.cursor = elements.length > 0 ? "pointer" : "default";
+          },
           plugins: {
             annotation: {
               annotations: {
@@ -158,7 +189,6 @@ export const DurationChart = ({
               },
               title: { align: "end", display: true, text: "Run After" },
             },
-
             y: {
               title: { align: "end", display: true, text: "Duration (seconds)" },
             },
