@@ -57,7 +57,8 @@ export const LimitedItemsList = ({
           // eslint-disable-next-line react/no-array-index-key
           <React.Fragment key={index}>
             <Text as="span">{item}</Text>
-            {index < displayItems.length - 1 || (shouldTruncate && remainingItems.length > 1) ? (
+            {index < displayItems.length - 1 ||
+            (shouldTruncate && remainingItems.length >= 1 && index === displayItems.length - 1) ? (
               <Text as="span">{separator}</Text>
             ) : undefined}
           </React.Fragment>
